/*
 * supporter.cpp
 *
 *  Created on: May 16, 2009
 *      Author: michael
 */


#include "supporter.h"
#include <iostream>
#include <algorithm>
#include <stdio.h>


const string Supporter::name     = "Supporter";
const string Supporter::version  = "v1.3.20100302";
const string Supporter::author   = "Michael Mampaey";



Supporter::Supporter()
{
	datafilename    = "";
	itemsetfilename = "";
	outfilename     = "";

	trans_count = 0;

	items = map< int, vector<uint>* >();
	
	all_tid_list = 0;
	
	return;

}



Supporter::~Supporter()
{
	cleanup();
	
	return;
}



void Supporter::cleanup()
{
	map<int, vector<uint>* >::iterator mvi_it;
	for (mvi_it=items.begin(); mvi_it!=items.end(); ++mvi_it) {
		delete mvi_it->second;
		mvi_it->second = 0;
	}
	items.clear();
	
	delete all_tid_list;
	
	return;
}



vector<int> *Supporter::read_line(FILE *file, bool allow_negative)
{
	if (file) {
		if (feof(file))
			return 0;
		vector<int> * itemset = new vector<int>();
		char c;
		bool skipline, firstchar = true;
		do {
			skipline = false;
			int item = 0, pos = 0, sign = 1;
			c = getc(file);
			if (firstchar && c == '#') {
				while (c != '\n' && c != '\r' && !feof(file))
					c = getc(file);
				skipline = true;
				continue;
			}
			else
				firstchar = false;
			if (c=='-' && allow_negative) {
				sign = -1;
				c = getc(file);
			}
			while((c >= '0') && (c <= '9')){
				item *= 10;
				item += c - '0';
				c = getc(file);
				++pos;
			}
			if (pos)
				itemset->push_back(sign * item);
		} while (skipline || (c != '\n' && c != '\r' && !feof(file)));

		if (feof(file) && itemset->size()==0) {
			delete itemset;
			return 0;
		}

		return itemset;
	}
	else {
		if (cin.eof())
			return 0;
		string itemset_line;
		getline(cin, itemset_line);
		uint ll = itemset_line.length();

		vector<int> *itemset = new vector<int>();
		char c;
		uint i = 0;
		do {
			int item = 0, pos = 0, sign = 1;
			c = itemset_line[i++];
			if (c=='-' && allow_negative) {
				sign = -1;
				c = itemset_line[i++];
			}
			while ((c >= '0') && (c <= '9')){
				item *= 10;
				item += c - '0';
				c = itemset_line[i++];
				++pos;
			}
			if (pos) {
				itemset->push_back(sign * item);
			}
		} while ( i < ll );

		return itemset;
	}
}



void Supporter::read_data()
{
	cleanup();

	FILE *infile = fopen(datafilename.c_str(), "rt");
	trans_count = 0;

	while (vector<int> *itemset = read_line(infile)) {
		++trans_count;
		vector<int>::iterator vit;
		for (vit=itemset->begin(); vit!=itemset->end(); ++vit) {
			if (items[*vit] == 0)
				items[*vit] = new vector<uint>();
			items[*vit]->push_back(trans_count);
		}
		delete itemset;
	}

	fclose(infile);

	make_all_tid_list();

	return;
}



void Supporter::make_all_tid_list()
{
	all_tid_list = new vector<uint>();
	for (uint i=1; i<=trans_count; ++i)
		all_tid_list->push_back(i);
	
	return;
}



void Supporter::run()
{
	FILE *setfile = 0;
	FILE *outfile = 0;
	if (itemsetfilename != "-")
		setfile = fopen(itemsetfilename.c_str(), "rt");
	if (outfilename != "-")
		outfile = fopen(outfilename.c_str(), "wt");

	while (vector<int> *itemset = read_line(setfile, true)) {
		uint support;
		vector<int>::iterator vit;
		map<int, vector<uint>* >::iterator loc;
		if (itemset->size() == 0) {
			support = trans_count;
		}
		else if (itemset->size() == 1) {
			loc = items.find(abs(*(itemset->begin())));
			if (loc == items.end())
				support = 0;
			else
				support = loc->second->size();
			if (*(itemset->begin()) < 0)
				support = trans_count - support;
		}
		else {
			vector<uint> *tid_list = 0, *tmp = 0;
			vit = itemset->begin();
			loc = items.find(abs(*vit));
			if (loc == items.end()) {
				if (*vit < 0) {
					tid_list = new vector<uint>(*all_tid_list);
					tmp = new vector<uint>();
					++vit;
				}
				else {
					tid_list = new vector<uint>();
					vit = itemset->end();
				}
			}
			else {
				if (*vit < 0) {
					tid_list = new vector<uint>();
					insert_iterator<vector<uint> > iit (*tid_list, tid_list->begin());
					set_difference(all_tid_list->begin(), all_tid_list->end(), loc->second->begin(), loc->second->end(), iit);
				}
				else {
					tid_list = new vector<uint>(*(loc->second));
				}
				tmp = new vector<uint>();
				++vit;
			}
			for ( ; vit!=itemset->end(); ++vit) {
				loc = items.find(abs(*vit));
				if (loc == items.end()) {
					if (*vit < 0) {
						continue;
					}
					else {
						tid_list->clear();
						break;
					}
				}
				tmp->clear();
				insert_iterator<vector<uint> > iit (*tmp, tmp->begin());
				if (*vit < 0)
					set_difference(tid_list->begin(), tid_list->end(), loc->second->begin(), loc->second->end(), iit);
				else
					set_intersection(tid_list->begin(), tid_list->end(), loc->second->begin(), loc->second->end(), iit);
				swap(tmp, tid_list);
			}
			support = tid_list->size();
			delete tmp;
			delete tid_list;
		}

		if (outfile) {
			for (vit=itemset->begin(); vit!=itemset->end(); ++vit) {
				fprintf(outfile, "%d ", *vit);
			}
			fprintf(outfile, "(%d)\n", support);
			if (!setfile)
				fflush(outfile);
		}
		else {
			for (vit=itemset->begin(); vit!=itemset->end(); ++vit) {
				cout << *vit << ' ';
			}
			cout << '(' << support << ')' << endl;
		}

		delete itemset;

	}
	
	if (setfile)
		fclose(setfile);
	if (outfile)
		fclose(outfile);

	return;
}



void Supporter::stats()
{
	uint one_count = 0;
	map<int, vector<uint>* >::iterator it;
	for (it = items.begin(); it != items.end(); ++it) {
		one_count += it->second->size();
	}

	cout << "transactions :\t" << trans_count << endl;
	cout << "items        :\t" << items.size() << endl;
	cout << "density      :\t" << (double)one_count / (trans_count * items.size()) << endl;

	return;

}



void Supporter::singletons()
{
	map<int, vector<uint>* >::iterator it;
	for (it = items.begin(); it != items.end(); ++it) {
		cout << it->first << " (" << it->second->size() << ")" << endl;
	}

	return;
}

